"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const util_1 = require("util");
const harmony_1 = require("@teambit/harmony");
const get_port_1 = __importDefault(require("get-port"));
const fs_extra_1 = __importDefault(require("fs-extra"));
const webpack_1 = __importDefault(require("webpack"));
const ui_aspect_1 = require("./ui.aspect");
const cli_1 = require("@teambit/cli");
const start_cmd_1 = require("./start.cmd");
const graphql_1 = require("@teambit/graphql");
const webpack_config_1 = __importDefault(require("./webpack/webpack.config"));
const exceptions_1 = require("./exceptions");
const create_root_1 = require("./create-root");
const utils_1 = require("bit-bin/dist/utils");
const express_1 = require("@teambit/express");
const component_1 = require("@teambit/component");
const ui_build_cmd_1 = require("./ui-build.cmd");
const ui_server_1 = require("./ui-server");
const logger_1 = require("@teambit/logger");
class UiMain {
    constructor(config, graphql, uiRootSlot, express, onStartSlot, componentExtension, logger) {
        this.config = config;
        this.graphql = graphql;
        this.uiRootSlot = uiRootSlot;
        this.express = express;
        this.onStartSlot = onStartSlot;
        this.componentExtension = componentExtension;
        this.logger = logger;
    }
    build(uiRootName) {
        return __awaiter(this, void 0, void 0, function* () {
            const [name, uiRoot] = this.getUi(uiRootName);
            const config = webpack_config_1.default(uiRoot.path, [yield this.generateRoot(uiRoot.extensionsPaths, uiRoot.aspectPaths, name)], name);
            const compiler = webpack_1.default(config);
            const compilerRun = util_1.promisify(compiler.run.bind(compiler));
            return compilerRun();
        });
    }
    createRuntime({ uiRootName, pattern, dev, port }) {
        return __awaiter(this, void 0, void 0, function* () {
            const [name, uiRoot] = this.getUi(uiRootName);
            this.componentExtension.setHostPriority(name);
            const uiServer = ui_server_1.UIServer.create({
                express: this.express,
                graphql: this.graphql,
                uiRoot,
                uiRootExtension: name,
                ui: this,
                logger: this.logger,
            });
            const targetPort = yield this.getPort(port);
            if (dev) {
                yield uiServer.dev({ port: targetPort });
            }
            else {
                yield uiServer.start({ port: targetPort });
            }
            if (uiRoot.postStart)
                uiRoot.postStart({ pattern });
            yield this.invokeOnStart();
            return uiServer;
        });
    }
    getPort(port) {
        return __awaiter(this, void 0, void 0, function* () {
            if (port)
                return port;
            return this.config.port || this.selectPort();
        });
    }
    registerOnStart(onStartFn) {
        this.onStartSlot.register(onStartFn);
        return this;
    }
    invokeOnStart() {
        return __awaiter(this, void 0, void 0, function* () {
            const promises = this.onStartSlot.values().map((fn) => fn());
            yield Promise.all(promises);
        });
    }
    registerUiRoot(uiRoot) {
        return this.uiRootSlot.register(uiRoot);
    }
    getUi(uiRootName) {
        if (uiRootName) {
            const root = this.getUiRootOrThrow(uiRootName);
            return [uiRootName, root];
        }
        const uis = this.uiRootSlot.toArray();
        if (uis.length === 1)
            return uis[0];
        const uiRoot = uis.find(([, root]) => root.priority);
        if (!uiRoot)
            throw new exceptions_1.UnknownUI('default');
        return uiRoot;
    }
    getUiRootOrThrow(uiRootName) {
        const uiSlot = this.uiRootSlot.get(uiRootName);
        if (!uiSlot)
            throw new exceptions_1.UnknownUI(uiRootName);
        return uiSlot;
    }
    createLink(extensionPaths, aspectPaths, rootExtensionName) {
        return create_root_1.createRoot(extensionPaths, aspectPaths, rootExtensionName);
    }
    generateRoot(extensionPaths, aspectPaths, rootExtensionName) {
        return __awaiter(this, void 0, void 0, function* () {
            const contents = yield create_root_1.createRoot(extensionPaths, aspectPaths, rootExtensionName);
            const filepath = path_1.resolve(path_1.join(__dirname, `ui.root${utils_1.sha1(contents)}.js`));
            if (fs_extra_1.default.existsSync(filepath))
                return filepath;
            fs_extra_1.default.outputFileSync(filepath, contents);
            return filepath;
        });
    }
    selectPort() {
        return __awaiter(this, void 0, void 0, function* () {
            const [from, to] = this.config.portRange;
            return get_port_1.default({ port: get_port_1.default.makeRange(from, to) });
        });
    }
    static provider([cli, graphql, express, componentExtension, loggerMain], config, [uiRootSlot, onStartSlot]) {
        return __awaiter(this, void 0, void 0, function* () {
            const logger = loggerMain.createLogger(ui_aspect_1.UIAspect.id);
            const ui = new UiMain(config, graphql, uiRootSlot, express, onStartSlot, componentExtension, logger);
            cli.register(new start_cmd_1.StartCmd(ui));
            cli.register(new ui_build_cmd_1.UIBuildCmd(ui));
            return ui;
        });
    }
}
exports.UiMain = UiMain;
UiMain.defaultConfig = {
    portRange: [3000, 3200],
};
UiMain.runtime = cli_1.MainRuntime;
UiMain.dependencies = [cli_1.CLIAspect, graphql_1.GraphqlAspect, express_1.ExpressAspect, component_1.ComponentAspect, logger_1.LoggerAspect];
UiMain.slots = [harmony_1.Slot.withType(), harmony_1.Slot.withType()];
ui_aspect_1.UIAspect.addRuntime(UiMain);
//# sourceMappingURL=ui.main.runtime.js.map